"""In this example we observe the overflow behavior of the Time Tagger. Overflows occur when the data
rate is too high and the buffer onboard the Time Tagger is completely filled. In this situation, data
loss occurs."""

import pathlib
import sys
from time import sleep
import numpy as np
from matplotlib import pyplot as plt
import TimeTagger

try:
    # We load the CustomCounter class that we will discuss in detail in example 4-C.
    # It is a sub-class of CustomMeasurement and resembles the standard Counter class,
    # but in comparison Counter it is able to track overflows.
    sys.path.append(str(pathlib.Path(__file__).parent.parent / "4-working-with-raw-time-tags"))
    from C_custom_measurement import CustomCounter
except ModuleNotFoundError as e:
    print(e)
    exit()

# Create a TimeTagger instance to control your hardware
tagger = TimeTagger.createTimeTagger()
# Call reset to set all settings to their default value
tagger.reset()

# Using the test signal, it is hardly possible on the Time Tagger Ultra to reach the overflow mode,
# so we take four channels in this case. For the Time Tagger 20, one channel is sufficient.
if tagger.getModel() == 'Time Tagger Ultra':
    CHANNELS = list(tagger.getChannelList(TimeTagger.TT_CHANNEL_RISING_EDGES)[:4])
else:
    CHANNELS = tagger.getChannelList(TimeTagger.TT_CHANNEL_RISING_EDGES)[:1]

# Activate test signal on the selected channels
tagger.setTestSignal(CHANNELS, True)

# We use the CustomCounter class that we loaded from example 4-C
counter = CustomCounter(tagger=tagger,
                        channels=CHANNELS,
                        binwidth=1000000000,
                        n_values=30000)

default_divider = tagger.getTestSignalDivider()
print("The default test signal divider is: {}".format(default_divider))

sleep(2)

# We increase the data rate successively by reducing the TestSignalDivider
divider = default_divider
while divider > 1 and not tagger.getOverflows():
    divider -= 1
    sys.stdout.write("divider = {:2}\r".format(divider))
    tagger.setTestSignalDivider(divider)
    sleep(.3)
print("\nOverflows occurred at test signal divider of {}".format(divider))

# We let the Time Tagger run for two more seconds in the overflow
sleep(2)

# Reset TestSignalDivider to the default value to recover from overflow mode
tagger.setTestSignalDivider(default_divider)
sleep(2)
counter.stop()

# Plot the result
fig = plt.figure()
plt.plot(counter.getIndex()/1E12, counter.getData(exclude_overflows=True))
plt.xlabel("Time, s")
plt.ylabel("Counts/s")
plt.annotate("""In the overflow mode,
there are gaps in curve. After the
overflown buffer is emptied by the
USB transfer, it can accumulate normal
time-tags for a short period, before
it overflows again. These time-tags are
displayed between the gaps.""",
             (26, np.max(counter.getData(exclude_overflows=False))),
             xytext=(-150, 0),
             textcoords='offset pixels',
             va='top',
             ha='right',
             arrowprops={'arrowstyle': '->'})
plt.show()
